# Constants used in the project

# Significance threshold defines a threshold for the significance of the feature for it to be considered at all
SIGNIFICANCE_THRESHOLD = 0.1
# Default number of top explanations to show
TOP_K_DEFAULT = 1
# Default number of figures in one row
DEFAULT_FIGS_IN_ROW = 2
